/*
 *  Licensed to the author under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package net.java.quickcheck;

import static java.lang.String.format;

import java.io.PrintWriter;

import net.java.quickcheck.util.Assert;

class RunnerImpl<T> extends AbstractRunner<T> {

	private int fMaxRuns;
	private int fCurrentRuns;
	private int fRuns;

	private PrintWriter fWriter;

	private Characteristic<T> fCharacteristic;

	private Generator<T> fGenerator;

	public RunnerImpl() {
		setRuns(QuickCheck.getDefaultNumberOfRuns());
	}

	public void forAll() {
		Assert.notNull(fCharacteristic, "characteristic");
		Assert.notNull(fGenerator, "generator");
		for (int count = 0; count < fRuns; count++) {
			try {
				doSetup();
				check(fGenerator);
			} finally {
				doTearDown();
			}
		}
	}

	private void check(Generator<T> generator) {

		Assert.notNull(generator, "generator");

		while (fCurrentRuns < fMaxRuns) {
			fCurrentRuns++;
			T value = generator.next();
			printCurrentValue(fCurrentRuns, value);
			try {
				fCharacteristic.specify(value);
				return;
			} catch (GuardException e) {
				print(format("%d: skipped [%s]", fCurrentRuns, value));
			} catch (Throwable e) {
				throwFailedException(value, e, fCharacteristic, fCurrentRuns);
			}
		}
		throw new GeneratorException(format(
				"Generator max tries (%s) reached. "
						+ "Check your guard expressions.", fMaxRuns), generator);
	}

	public static int getMaxGeneratorTries(int maxRuns) {
		return maxRuns * 10;
	}

	private void doTearDown() {
		try {
			fCharacteristic.tearDown();
		} catch (Exception e) {
			throw new CharacteristicException("teardown failed", e,
					fCharacteristic);
		}
	}

	private void doSetup() {
		try {
			fCharacteristic.setUp();
		} catch (Exception e) {
			throw new CharacteristicException("setup failed", e,
					fCharacteristic);
		}
	}

	public PrintWriter getWriter() {
		return fWriter;
	}

	public void setWriter(PrintWriter writer) {
		fWriter = writer;
	}

	public int getRuns() {
		return fRuns;
	}

	public void setRuns(int runs) {
		Assert.greaterOrEqual(QuickCheck.MIN_NUMBER_OF_RUNS, runs, "runs");
		fMaxRuns = 10 * runs;
		fRuns = runs;
	}

	public Characteristic<T> getCharacteristic() {
		return fCharacteristic;
	}

	public void setCharacteristic(Characteristic<T> characteristic) {
		fCharacteristic = characteristic;
	}

	public Generator<T> getGenerator() {
		return fGenerator;
	}

	public void setGenerator(Generator<T> generator) {
		fGenerator = generator;
	}
}